import sys
import csv
import pickle

# para0 : optional -d for debugging
# para1 : path to the fragment_bac file
# para2 : path to the bac_pool file
# para3 : path to the pool_unigene file
# para4 : path to the input bac unigene file, the deconvolution from the first step
# para5 : path to the output data file for LP
# para6 : path to the output pickle file

DEBUGGING = False
if sys.argv[1] == '-d':
	DEBUGGING = True
	del sys.argv[1]

#declare some variables

#This set is from the fragment_bac file
fragments = set() #the set of fragment ids. 
#This set is from the fragment_bac file
bacs = set() #the set of bac ids. 
#This set is from the pool_unigne file
pools = set() #the set of pool ids. 
#This set is from the pool_unigene file
unigenes = set() #the set of unigene ids

#This dictionary is from the fragment_bac file
fragment_2_bacs = {} #a dictionary of bac set indexed by fragment id
#This dictionary is from the bac_pool file
bac_2_pools = {} #a dictionary of pool set indexed by bac_id
#This dictionary is read from the pool unigene file
pool_2_unigenes = {} #a dictionary of unigene set indexed by pool_id 
#This dictionary is read from the bac unigene file
bac_2_unigenes_step1 = [] # a list of (bac_id, unigene set) pairs that are produced by our naive deconvolution algorithm


#temporary tables used in the course of deconvolution
fragment_2_pools = {} # a dictionary of pool set indexed by fragment_id
bac_2_fragments = {}# a dictionary of fragment set indexed by bac_id
pool_2_fragments = {} # a dictionary of fragment set indexed by pool_id
unigene_2_pools = {} # a dictionary of pool set indexed by unigene_id

#The output of the algorithm
unigene_2_fragments = {} #a dictionary of fragment set indexed by unigene_id
fragment_2_unigenes = {} #a dictionary of unigene set indexed by fragment_id


#read in the fragment bac table
input_file = open(sys.argv[1])
one_line = input_file.readline()
while one_line:
	one_line = one_line.strip()
	records = one_line.split(',')
	fragment_id = records[0]
	current_bac_set = set(records[1:])
	fragments.add(fragment_id)
	bacs = bacs | current_bac_set
	fragment_2_bacs[fragment_id] = current_bac_set
	one_line = input_file.readline()
input_file.close()

if DEBUGGING:
	print len(fragments)
	print fragments
	print len(bacs)
	print bacs
	print len(fragment_2_bacs)
	for ii in fragments:
		print ii, fragment_2_bacs[ii]

#pivot the fragment_2_bacs table to produce the bac_2_fragments table
for ii in bacs:
	bac_2_fragments[ii] = set()
for jj in fragments:
	current_bac_set = fragment_2_bacs[jj]
	for bac_id in current_bac_set:
		bac_2_fragments[bac_id].add(jj)

if DEBUGGING:
	print len(bac_2_fragments)
	for ii in bacs:
		print ii, bac_2_fragments[ii]

#read in the bac pool table
input_file = open(sys.argv[2])
one_line = input_file.readline()
while one_line:
	one_line = one_line.strip()
	records = one_line.split(',')
	bac_id = records[0]
	current_pool_set = set(records[1:])
	bac_2_pools[bac_id] = current_pool_set
	one_line = input_file.readline()
input_file.close()

if DEBUGGING:
	print len(bac_2_pools)
	for ii in bac_2_pools.keys():
		print ii, bac_2_pools[ii]

#read in the pool unigene table 
input_file = open(sys.argv[3])
one_line = input_file.readline()
while one_line:
	one_line = one_line.strip()
	records = one_line.split(',')
	pool_id = records[0]
	pools.add(pool_id)
	current_unigene_set = set(records[1].split('|'))
	unigenes = unigenes | current_unigene_set
	pool_2_unigenes[pool_id] = current_unigene_set
	one_line = input_file.readline()
input_file.close()

if DEBUGGING:
	print len(pools)
	print pools
	for ii in pools:
		print ii, pool_2_unigenes[ii]

#read in the bac unigene file output from our naive deconvolution algorithm
input_file = open(sys.argv[4])
one_line = input_file.readline()
while one_line:
	one_line = one_line.strip()
	records = one_line.split(',')
	bac_id = records[0]
	current_unigene_set = set(records[2:])
	bac_2_unigenes_step1.append((bac_id, current_unigene_set))
	one_line = input_file.readline()

if DEBUGGING:
	for ii in bac_2_unigenes_step1:
		print ii


#############################################
#end reading the input files
#############################################

###############################################
#start the deconvolution process
###############################################
#initialization
for ii in fragments:
	fragment_2_pools[ii] = set()

#Step 1:, use the pooling provided by overlapping of BACs to deconvolute the relationships between fragments and probe pools
for ii in fragments:
	current_bac_set = fragment_2_bacs[ii]
	hybri_pools = pools
	for bac_id in current_bac_set:
		try:
			positive_set = bac_2_pools[bac_id]
		except:
			positive_set = set()
		hybri_pools = hybri_pools & positive_set
	fragment_2_pools[ii] = hybri_pools

#print out for debugging

if DEBUGGING:
	print len(fragment_2_pools)
	count = 0 
	for ii in fragments:
		print ii, fragment_2_pools[ii]
		if len(fragment_2_pools[ii]) > 1:
			count = count + 1
	print count

#Step 2:, use the pooling of the probes to deconvolute the relationships between fragments and unigenes
#pivot the fragment_2_pools table
#initialize
for ii in pools:
	pool_2_fragments[ii] = set()
	
for jj in fragments:
	current_pool_set = fragment_2_pools[jj]
	for pool_id in current_pool_set:
		pool_2_fragments[pool_id].add(jj)

if DEBUGGING:
	for ii in pools:
		print ii, pool_2_fragments[ii]

#pivot the pool_2_unigenes table
#initialize
for ii in unigenes:
	unigene_2_pools[ii]= set()

for jj in pools:
	current_unigene_set = pool_2_unigenes[jj]
	for unigene_id in current_unigene_set:	
		unigene_2_pools[unigene_id].add(jj)

if DEBUGGING:
	for ii in unigenes:
		print ii, unigene_2_pools[ii]

for ii in unigenes:
	hybri_fragments = fragments
	pools_set = unigene_2_pools[ii]
	for pool_id in pools_set:
		try:
			positive_fragment_set = pool_2_fragments[pool_id]
		except:
			positive_fragment_set = set()
		hybri_fragments = hybri_fragments & positive_fragment_set
	unigene_2_fragments[ii] = hybri_fragments

#pivot the unigene_2_fragments table to produce the fragment_2_unigenes table
#initialization
for ii in fragments:
	fragment_2_unigenes[ii] = set()
for jj in unigenes:
	current_fragment_set = unigene_2_fragments[jj]
	for fragment_id in current_fragment_set:
		fragment_2_unigenes[fragment_id].add(jj)

if DEBUGGING:
	count = 0 
	for ii in fragments:
		print ii, fragment_2_unigenes[ii]
		count = count + len(fragment_2_unigenes[ii])
	print count

#################################################
#Now this part is to generate the data file for LP
#################################################
output_file = open(sys.argv[5],'w')
output_file.write('data;\n')
output_file.write('param num_of_fragments := %d;\n'%len(fragments))
output_file.write('param num_of_unigenes := %d;\n'%len(unigenes))
output_file.write('param num_of_constraints := %d;\n'%len(bac_2_unigenes_step1))
output_file.write('set fragments:= ')
for ii in fragments:
	output_file.write(' %s'%ii)
output_file.write(';\n')

output_file.write('set unigenes:= ')
for ii in unigenes:
	output_file.write(' %s'%ii)
output_file.write(';\n')

output_file.write('set edges:= ')
for ii in fragments:
	if len(fragment_2_unigenes[ii]) > 0:
		output_file.write('\n\t(%s,*)'%ii)
		current_unigene_set = fragment_2_unigenes[ii]
		for jj in current_unigene_set:
			output_file.write(' %s'%jj)
output_file.write(';\n')

for ii in range(len(bac_2_unigenes_step1)):
	output_file.write('set constraints[%d] :='%(ii+1))
	bac_id = bac_2_unigenes_step1[ii][0]
	current_fragment_set = bac_2_fragments[bac_id]
	current_unigene_set = bac_2_unigenes_step1[ii][1]
	for fragment_id in current_fragment_set:
		positive_unigene_set = fragment_2_unigenes[fragment_id] & current_unigene_set
		if len(positive_unigene_set) > 0:
			output_file.write('\n\t(%s,*)'%fragment_id)
			for uu in positive_unigene_set:
				output_file.write(' %s'%uu)
	output_file.write(';\n')
output_file.write('end;\n')

###############################################################################
#The last part is to generate the pickle file
###############################################################################
	
class dummy_class:
	pass
	
to_be_pickled = dummy_class()


to_be_pickled.fragments = fragments #a set object
to_be_pickled.bacs = bacs #a set object
to_be_pickled.unigenes = unigenes # a set object 
to_be_pickled.fragment_2_bacs = fragment_2_bacs # a dictionary referenced by fragment_id
to_be_pickled.bac_2_unigenes_step1 = bac_2_unigenes_step1 #a dictionary referenced by bac_id

pickle.dump(to_be_pickled, open(sys.argv[6],'w'),pickle.HIGHEST_PROTOCOL)




