import pickle
import csv
import sys
import re
import random

DEBUGGING = False
if sys.argv[1]=='-d':
	DEBUGGING = True
	del sys.argv[1]


pickle_file_path = sys.argv[1]
opt_sol_path = sys.argv[2]
constraints_file = sys.argv[3]
output_file_path = sys.argv[4]
number_of_trials = int(sys.argv[5])


class dummy_class:
	pass


#the first step is to read in the pickle file
agglomerate =  pickle.load(open(pickle_file_path))

bacs = agglomerate.bacs # a set object
unigenes = agglomerate.unigenes  # a set object 
contigs = agglomerate.contigs # a list object
contig_2_bacs = agglomerate.contig_2_bacs # a dictionary referenced by contigs
bac_2_unigene_edges = agglomerate.bac_2_unigene_edges # a set of pairs (bac_id, unigene_id)
print 'total number of bacs:', len(bacs)
print 'total number of unigenes:', len(unigenes)
print 'total number of contigs:', len(contigs)
print 'total number of bac_2_unigene_edges:', len(bac_2_unigene_edges)

unigene_2_contigs = {} # a dictionary indexed by unigenes. Each entry is a list of contigs that the unigene might assign to.
for unigene_id in unigenes:
	unigene_2_contigs[unigene_id] = []

bac_2_unigenes_step1 = [] 
#read in the constraints file
input_file = open(constraints_file)
one_line = input_file.readline()
while one_line:
	one_line = one_line.strip()
	records = one_line.split(',')
	bac_id = records[0]
	current_unigene_set = set(records[2:])
	bac_2_unigenes_step1.append((bac_id, current_unigene_set))
	one_line = input_file.readline()
if DEBUGGING:
	for ii in bac_2_unigenes_step1:
		print ii


#the second step is to read in the optimal solution
#ctg_unigene_assign_re = re.compile(r'x\[(ctg\d+),(\d+)\] = (1|0.\d+)')
ctg_unigene_assign_re = re.compile(r'x\[(\w+),((?:S\d+)|(?:g\d+.\d))\] = ((?:0.\d+)|(?:\d.\d+e-\d+)|(?:1))') #customized for evaluation of rice

opt_sol_file = open(opt_sol_path)
one_line = opt_sol_file.readline()
while one_line:
	try:
		records = ctg_unigene_assign_re.findall(one_line)[0]	
		unigene_2_contigs[records[1]].append([records[0],float(records[2])])	
	except:
		print one_line
	one_line = opt_sol_file.readline()
opt_sol_file.close()


previous_number_of_satisfied_constraints = 0
while number_of_trials > 0:

	number_of_trials  = number_of_trials - 1
	bac_2_unigenes = {} # a dictionary indexed by bacs. Each entry is a list of unigenes the corresponding bac is assigned to.
	for bac_id in bacs:
		bac_2_unigenes[bac_id] = set()

	unigene_2_contig_final_assign = {} # a dictionary indexed by unigenes. Each entry the contig that the unigene is assigned to.

	#the third step is to round the solution
	for unigene_id in unigenes:
		random_float = random.random()
		sum_now = 0
		choice = -1
		while (random_float > sum_now):
			choice = choice + 1
			if choice == len(unigene_2_contigs[unigene_id]):
				choice = -1
				break
			else:
				sum_now = sum_now + unigene_2_contigs[unigene_id][choice][1]
		if choice == -1:
			print 'null assignment %s'%unigene_id
			unigene_2_contig_final_assign[unigene_id] = ''
		else:
			unigene_2_contig_final_assign[unigene_id] = unigene_2_contigs[unigene_id][choice][0]

	for unigene_id in unigenes:
		contig_id = unigene_2_contig_final_assign[unigene_id]
		if contig_id <> '':
			current_bacs = contig_2_bacs[contig_id]
			for bac_id in current_bacs:
				if (bac_id, unigene_id) in bac_2_unigene_edges:
					bac_2_unigenes[bac_id].add(unigene_id)

	#calculate the number of satisfied constraints
	number_of_satisfied_constraints = 0
	for ii in bac_2_unigenes_step1:
		bac_id = ii[0]
		if bac_id in bacs: #some bacs are singletons, so they are not counted
			if len( ii[1] & bac_2_unigenes[bac_id]) > 0:
				number_of_satisfied_constraints = number_of_satisfied_constraints + 1
	print 'the number of satisfied constraints at traial %d: %d'%(number_of_trials, number_of_satisfied_constraints)

	#the fouth step to output the optimal solution
	if number_of_satisfied_constraints > previous_number_of_satisfied_constraints:
		previous_number_of_satisfied_constraints = number_of_satisfied_constraints
		print 'output a new file'
		output_file = open(output_file_path,'w')
		for bac_id in bacs:
			current_unigenes_set = bac_2_unigenes[bac_id]
			for unigene_id in current_unigenes_set:
				output_file.write(unigene_id)
				output_file.write(',')
				output_file.write(bac_id)
				output_file.write('\n')
		output_file.close()