import pickle
import csv
import sys
import re
import random

DEBUGGING = False
if sys.argv[1]=='-d':
	DEBUGGING = True
	del sys.argv[1]

pickle_file_path = sys.argv[1]
opt_sol_path = sys.argv[2]
output_file_path = sys.argv[3]
number_of_trials = int(sys.argv[4])



class dummy_class:
	pass


#the first step is to read in the pickle file
agglomerate =  pickle.load(open(pickle_file_path))

bacs = agglomerate.bacs # a set object
unigenes = agglomerate.unigenes  # a set object 
fragments = agglomerate.fragments # a set object
fragment_2_bacs = agglomerate.fragment_2_bacs # a dictionary referenced by fragment_id
bac_2_unigenes_step1 = agglomerate.bac_2_unigenes_step1 #a dictionary referenced by bac_id

print 'total number of bacs:', len(bacs)
print 'total number of unigenes:', len(unigenes)
print 'total number of fragments:', len(fragments)
print 'total number of fragment_2_bacs:', len(fragment_2_bacs)
print 'total number of constraints:', len(bac_2_unigenes_step1)


unigene_2_fragments = {} # a dictionary indexed by unigenes. Each entry is a list of fragments that the unigene might assign to.
for unigene_id in unigenes:
	unigene_2_fragments[unigene_id] = []



#the second step is to read in the optimal solution
#ctg_unigene_assign_re = re.compile(r'x\[(\w+),(g\d+.\d)\] = ((0.\d+)|(\d.\d+e-\d+)|(1))') #customized for evaluation of rice
#fragment_unigene_assign_re = re.compile(r'x\[(\w+),(S\d+)\] = ((0.\d+)|(\d.\d+e-\d+)|(1))') #customized for evaluation of rice
fragment_unigene_assign_re = re.compile(r'x\[(\w+),((?:S\d+)|(?:g\d+.\d))\] = ((?:0.\d+)|(?:\d.\d+e-\d+)|(?:1))') #customized for evaluation of rice

opt_sol_file = open(opt_sol_path)
one_line = opt_sol_file.readline()
while one_line:
	try:
		records = fragment_unigene_assign_re.findall(one_line)[0]	
		unigene_2_fragments[records[1]].append([records[0],float(records[2])])	
	except:
		print one_line
	one_line = opt_sol_file.readline()
opt_sol_file.close()

#the third step is to round the solution

previous_number_of_satisfied_constraints = 0
while number_of_trials > 0:
	number_of_trials  = number_of_trials - 1
	
	unigene_2_fragment_final_assign = {} # a dictionary indexed by unigenes. Each entry the fragment that the unigene is assigned to.
	unigene_2_bacs = {} # a dictionary indexed by unigenes. Each entry is a list of bacs that the corresponding unigene is assigned to.
	for unigene_id in unigenes:
		random_float = random.random()
		sum_now = 0
		choice = -1
		while (random_float > sum_now):
			choice = choice + 1
			if choice == len(unigene_2_fragments[unigene_id]):
				choice = -1
				break
			else:
				sum_now = sum_now + unigene_2_fragments[unigene_id][choice][1]
		if choice == -1:
			print 'null assignment:%s'%unigene_id
			unigene_2_fragment_final_assign[unigene_id] = '';
		else:
			unigene_2_fragment_final_assign[unigene_id] = unigene_2_fragments[unigene_id][choice][0]

	for unigene_id in unigenes:
		fragment_id = unigene_2_fragment_final_assign[unigene_id]
		if fragment_id <> '':
			current_bac_set = fragment_2_bacs[fragment_id]
			unigene_2_bacs[unigene_id]= current_bac_set
		else:
			unigene_2_bacs[unigene_id]= set()

	#pivot the unigene_2_bacs table to produce the bac_2_unigenes table
	if DEBUGGING:
		print bacs
		
	bac_2_unigenes = {}
	for bac_id in bacs:
		bac_2_unigenes[bac_id] = set()
	for unigene_id in unigenes:
		current_bac_set = unigene_2_bacs[unigene_id]
		for bac_id in current_bac_set:
			bac_2_unigenes[bac_id].add(unigene_id)
	#calculate the number of satisfied constraints
	number_of_satisfied_constraints = 0
	for ii in bac_2_unigenes_step1:
		bac_id = ii[0]
		if len( ii[1] & bac_2_unigenes[bac_id]) > 0:
			number_of_satisfied_constraints = number_of_satisfied_constraints + 1
	print 'the number of satisfied constraints at traial %d: %d'%(number_of_trials, number_of_satisfied_constraints)
	#the fouth step to output the optimal solution
	if number_of_satisfied_constraints > previous_number_of_satisfied_constraints:
		previous_number_of_satisfied_constraints = number_of_satisfied_constraints
		print 'output a new file'
		output_file = open(output_file_path,'w')
		for unigene_id in unigenes:
			current_bac_set = unigene_2_bacs[unigene_id]
			for bac_id in current_bac_set:
				output_file.write(unigene_id)
				output_file.write(',')
				output_file.write(bac_id)
				output_file.write('\n')
		#	print unigene_id, len(bacs)
		output_file.close()